/* -*- c-basic-offset: 2 -*- */
package nwnintf;

import java.io.*;
import java.util.*;

import nwnintf.format.tlk.*;
import nwnintf.format.mod.*;
import nwnintf.format.itp.*;
import nwnintf.format.itp.element.*;
import nwnintf.palette.*;

/** a python itpfile.py::ModuleFile and itpfile.py::DialogTlkFile
 * instances for use from java.
 */
public class NwnModule {

  TlkFile tlkfile_;
  ModFile modfile_;
  MultiPaletteTree mptree_;
  SezFile sezfile_;
  SavedScripts savedscripts_;

  public NwnModule( File modfilename ) throws Exception {

    File nwnroot = null;
    try {
      if( new File("scriptease.ini").canRead() ) {
        IniParser ip = new IniParser(new File("scriptease.ini"));
        nwnroot = new File( ip.get("Locations", "NWN") );
      }
    } catch( Exception e ) {
    }
    if( nwnroot == null ) {
      nwnroot = new File("C:\\NeverwinterNights\\NWN");
    }

    init( modfilename, nwnroot );
  }

  public NwnModule( File modfilename, File nwnroot ) throws Exception {
    init( modfilename, nwnroot );
  }

  private void init( File modfilename, File nwnroot ) throws Exception {
    modfile_ = new ModFile( modfilename );

    /* Open dialog.tlk in NWN install directory.
     */
    File dialogtlkfile;
    try {
      dialogtlkfile = new File( nwnroot, "dialog.tlk" );

      if( dialogtlkfile.canRead() ) {
	tlkfile_ = new TlkFile( dialogtlkfile );
      }
    } catch( Exception e ) {}


    if( tlkfile_ == null ) {
      // Make a stub tlkfile, it will just return string index numbers.
      tlkfile_ = new TlkFile();
      System.out.println("Can't find dialog.tlk");
    }

    /* Cache blueprint tree.
     */
    mptree_ = new MultiPaletteTree( modfile_, tlkfile_ );
    

    /* Load up the .sez file that records everything we have done to the
     * module.
     */
    sezfile_ = new SezFile(this);

    savedscripts_ = new SavedScripts( this );
  }

  /** Install a script resource in the module. Will silently replace existing
   * script by the same name.
   * @param resref The ResRef, not including the extension.
   * @param scriptstring The source code portion of the resource.
   * @param bytecodestring The compiled bytecode of the source code.
   */
  public void addScriptResource(ResRef resref, String scriptstring,
			    byte[] bytecode) throws IOException {
    
    modfile_.addupdateresource( new ResRef(resref, ResRef.tiNSS),
				scriptstring.getBytes() );
    modfile_.addupdateresource( new ResRef(resref, ResRef.tiNCS),
				bytecode );
  }


  /**
   * Responsible for actually writing the changes to the module file.
   * @see ScriptInstaller#attachScript(ResRef,String,ResRef)
   */
  public void attachScript(ResRef blueprint, String slotname,
			   ResRef scriptres ) throws Exception {

    ItpFile itpfile = new ItpFile( modfile_.resource( blueprint ) );
    ItpEntry root = itpfile.root();
    ItpResrefElt resref = (ItpResrefElt)root.element(slotname);
    resref.setValue( scriptres.name() );

    modfile_.addupdateresource( blueprint,
				new ItpWriter( itpfile ).bytes() );
  }


  /*
  public PlacedEntity[] placedEntitiesInModule() throws Exception {
    throw new Exception("placedEntitiesInModule() is not implemented.");
  }
  */

  /**
   * Return a list of all blueprints in this module. takes the list from
   * the palettes, rather than iterating over all blueprints, because
   * we need information from the palettes anyway.
   */
  public ModuleBlueprint[] blueprintsInModule() {

    Vector v = new Vector();

    /* Iterate over the cached palette trees.
     */
    for( Iterator it = mptree_.entries().iterator();
    it.hasNext(); ) {

      Map.Entry ment = (Map.Entry)it.next();
      String restype = (String)ment.getKey();
      PaletteTree ptree = (PaletteTree)ment.getValue();

      /* For each tree, iterate over leaf nodes.
       */
      for( Iterator it2 = ptree.blueprintnodes().iterator(); it2.hasNext(); ) {

        Map.Entry bpent = (Map.Entry)it2.next();
        ResRef resref = (ResRef)bpent.getKey();
        PaletteNode pnode = (PaletteNode)bpent.getValue();

        ModuleBlueprint mb =
          new ModuleBlueprint( resref,
            //new ResRef(resname, restype), 
            ((ItpStringElt)pnode.entry().element("NAME")).value()
          );
        v.add( mb );
      }
    }

    ModuleBlueprint[] blueprints = new ModuleBlueprint[ v.size() ];

    for( int i = 0; i < blueprints.length; ++i ) {
      blueprints[i] = (ModuleBlueprint)v.elementAt(i);
    }

    return blueprints;
  }




  public ModFile getModfile() {
    return modfile_;
  }

  /**
   * @see getTlkfile
   */
  public TlkFile getDialogTlkFile() {
    return tlkfile_;
  }


  public TlkFile getTlkfile() {
    return tlkfile_;
  }
  

  /**
   * Return a cached representation (palettepicker.py::MultiPaletteTree)
   * of the custom palettes so that blueprint picker dialogs can come up
   * faster.
   */
  public MultiPaletteTree getMultiPaletteTree() {
    return mptree_;
  }


  public SavedScripts getSavedScripts() {
    return savedscripts_;
  }

  public SezFile getSezFile() {
    return sezfile_;
  }

  /**
   * Save itp_, which represents the .sez file.
   */
  public void saveSezFile() throws Exception {
    sezfile_.save();

  }

  public static void main(String[] args) throws Exception {
    NwnModule nwnm = new NwnModule(new File("T0.mod"));
    nwnm.addScriptResource(new ResRef("testscript",-1),
			   "source",
			   "bytecode".getBytes());

    nwnm.attachScript(ResRef.parseFilename("shadowdoor.utd"), 
		      "OnClosed", new ResRef("testscript",-1));

    System.out.println(nwnm.blueprintsInModule());
  }

}
