/* -*- c-basic-offset: 2 -*- */

package nwnintf;
import java.io.*;
import nwnintf.format.itp.*;
import nwnintf.format.itp.element.*;
import nwnintf.format.mod.ModFile;
import nwnintf.format.mod.ResRef;

/**
 * Encapsulate a BlueSlotRef and the ItpEntry its information is stored in.
 */

public class SavedBlueSlot implements SavedSlot {

  public static String ORIGELT = "OrigScript";
  public static String GENELT =  "GenScript";
  public static String RESREFELT  =  "ResRef";
  public static String RESTYPEELT =  "ResType";
  public static String SLOTELT =  "Slot";


  BlueSlotRef slotref_;
  ItpEntry entry_;

  public SavedBlueSlot(ItpEntry ent) throws Exception {
    entry_ = ent;

    ResRef resref = 
      new ResRef( ((ItpResrefElt)ent.element(RESREFELT)).value(),
		  ((ItpInt32Elt)ent.element(RESTYPEELT)).value() );

    slotref_ =
      new BlueSlotRef( resref, 
		       ((ItpStringElt)ent.element(SLOTELT)).value(), false );
  }

  public ScriptSlotRef getSlotref() {
    return slotref_;
  }
  public ItpEntry getEntry() {
    return entry_;
  }


  public static SavedBlueSlot saveNew(BlueSlotRef blueslot, ModFile mod,
				      SezFile sez, ResRef script)
    throws Exception {

    ItpFile itpblue =
      new ItpFile(mod.resource(blueslot.getResref()));

    ItpResrefElt slotelt =
      (ItpResrefElt)(itpblue.root().element(blueslot.getSlot()));

    /* Save the old slot contents. */
    String origname = slotelt.value();

    slotelt.setValue( script.name() );
    mod.addupdateresource( blueslot.getResref(),
			   new ItpWriter(itpblue).bytes() );

    /* Record the existence of the scripts and where they're going.
     */
    ItpEntry ent = sez.newReslistEntry();

    ItpResrefElt orig = new ItpResrefElt(sez.itp());
    orig.setName(ORIGELT);
    orig.setValue(origname);

    ItpResrefElt gen = new ItpResrefElt(sez.itp());
    gen.setName(GENELT);
    gen.setValue(script.name());

    ItpResrefElt blueref = new ItpResrefElt(sez.itp());
    blueref.setName(RESREFELT);
    blueref.setValue(blueslot.getResref().name());

    ItpInt32Elt bluetyperef = new ItpInt32Elt(sez.itp());
    bluetyperef.setName(RESTYPEELT);
    bluetyperef.setValue(blueslot.getResref().typeInt());

    ItpStringElt slotref = new ItpStringElt(sez.itp());
    slotref.setName(SLOTELT);
    slotref.setValue(blueslot.getSlot());

    sez.itp().addElement(orig); ent.addElement(orig);
    sez.itp().addElement(gen); ent.addElement(gen);
    sez.itp().addElement(blueref); ent.addElement(blueref);
    sez.itp().addElement(bluetyperef); ent.addElement(bluetyperef);
    sez.itp().addElement(slotref); ent.addElement(slotref);
    
    return new SavedBlueSlot(ent);
  }

  /**
   * An existing saved slot has new text for the script.
   */
  public void update( ModFile mod, SezFile sez,
		      ScriptCode scriptcode ) throws Exception {
    ItpResrefElt elt = (ItpResrefElt)entry_.element(GENELT);
    mod.addupdateresource( new ResRef( elt.value(), ResRef.tiNSS ),
			   scriptcode.getSourcecode().getBytes() );
    mod.addupdateresource( new ResRef( elt.value(), ResRef.tiNCS ),
			   scriptcode.getBytecode() );
    
  }

  /**
   * Restore the slot of the blueprint this is saved into.
   * @return Resref (typeless) of the script that was originally
   * in this slot.
   */
  public ResRef delete( ModFile mod, SezFile sez ) throws Exception {

    String genscriptbase= ((ItpResrefElt)entry_.element(GENELT)).value();
    String orgscriptbase=((ItpResrefElt)entry_.element(ORIGELT)).value();

    /* Get rid of old generated script resources.
       */
    mod.deleteresource(new ResRef(genscriptbase, ResRef.tiNSS));
    mod.deleteresource(new ResRef(genscriptbase, ResRef.tiNCS));

    /* Restore the blueprint's script slot to its original form.
     */

    String bpresname = ((ItpResrefElt)entry_.element(RESREFELT)).value();
    int bprestype = ((ItpInt32Elt)entry_.element(RESTYPEELT)).value();

    byte[] blueprintdata = 
      mod.resource( new ResRef(bpresname, bprestype) );

    ItpFile blueitp = new ItpFile( blueprintdata );
    String slot = ((ItpStringElt)entry_.element(SLOTELT)).value();
    ItpResrefElt resref = (ItpResrefElt)(blueitp.root().element(slot));
    resref.setValue( orgscriptbase );

    /* And write that blueprint back to the module.
     */
    byte[] newdata = new ItpWriter( blueitp ).bytes();

    mod.addupdateresource( new ResRef(bpresname, bprestype ), newdata );


    /* Delete the entry for this script slot from the ScriptEase itp.
       */
    sez.removeFromReslist( entry_ );
      
    return new ResRef(orgscriptbase, -1);
  }

}
