/* -*- c-basic-offset: 2 -*- */

package nwnintf;
import java.io.*;
import java.util.*;
import nwnintf.format.*;
import nwnintf.format.mod.*;
import nwnintf.format.itp.*;
import nwnintf.format.itp.element.*;


/**
 * Represent and maintain the list of blueprint slots and scripts that
 * have been inserted by scriptease into the module. Sit on top of
 * the .sez ItpFile, particulary the <root>.ResList element.
 */

public class SavedScripts {

  NwnModule nwnmodule_;
  SezFile sezfile_;

  public SavedScripts(NwnModule nwn) {
    nwnmodule_ = nwn;
    sezfile_ = nwn.getSezFile();
  }


  /**
   * Get all existing SavedBlueSlots in the file and map
   * their BlueSlotRefs to them.
   * Map BlueSlotRef => SavedBlueSlot
   */
  private Map summarizeSavedBlueSlots() throws Exception {
    Map hm = new HashMap();
    for( int i = 0; i < sezfile_.reslist().length(); ++i ) {
      ItpEntry ent = sezfile_.reslist().entry(i);

      SavedBlueSlot sbs = new SavedBlueSlot(ent);
      hm.put( sbs.getSlotref(), sbs );
    }
    return hm;
  }

  /**
   * Set the ScriptSlots in the module to these ones.
   * @return A collection of SlotValue describing the exact changes
   * made to all blueprints and other scriptable objects (like dialog nodes).
   */
  public Collection save( Collection scriptslots ) throws Exception {

    /* Construct three sets.
     *
     * 1. Script slots in the previous saved version but not the new one.
     * 2. Script slots in the new saved version but not the old one.
     * 3. Script slots in both old and new.
     */

    /* To that end, all old ones. BlueSlotRef->SavedBlueSlot
     */
    Map oldsaved = summarizeSavedBlueSlots();
    //System.out.println("# old scripts: " + Integer.toString(oldsaved.size()));

    /* All new ones, keyed on the same ScriptSlotRef->ScriptSlot.
     */
    Map newsaved = new HashMap();
    for( Iterator it = scriptslots.iterator();
	 it.hasNext(); ) {
      ScriptSlot ss = (ScriptSlot)it.next();
      newsaved.put(ss.getSlotref(),ss);
    }
    //System.out.println("# new scripts: " + Integer.toString(newsaved.size()));


    /* Contains the applicable keys, not values, from the above maps. */
    Collection onlyold = new Vector();
    Collection onlynew = new Vector();
    Collection both = new Vector();

    /* Look at the old entries to see which ones are also in the new.
     */
    for( Iterator it = oldsaved.entrySet().iterator(); it.hasNext(); ) {
      ScriptSlotRef key = (ScriptSlotRef)(((Map.Entry)it.next()).getKey());
      if( newsaved.containsKey(key) ) {
	both.add( key );
      } else {
	onlyold.add( key );
      }
    }


    /* "both" and "onlyold" are both filled in. See which entries of new
     * are only in new.
     */
    for( Iterator it = newsaved.entrySet().iterator(); it.hasNext(); ) {
      ScriptSlotRef key = (ScriptSlotRef)(((Map.Entry)it.next()).getKey());
      if( oldsaved.containsKey(key) ) {
	// Don't add to both, that would have been done in previous loop.
      } else {
	onlynew.add( key );
      }
    }

    /* Now we can handle each class separately. */

    Collection changes = new HashSet();

    changes.addAll( saveOnlyOld( onlyold, oldsaved ) );
    changes.addAll( saveOnlyNew( onlynew, newsaved ) );
    changes.addAll( saveBoth( both, oldsaved, newsaved ) );

    /* And finally, save the .SEZ file. */
    nwnmodule_.saveSezFile();

    return changes;
  }


  /**
   * Restore old script slots to the way they were before, and
   * delete the associated resources we've added.
   *
   * @param oldkeys ScriptSlotRef keys that index into oldmap
   * @param values Contains instances of ItpEntry.
   * @return Collection of SlotValue describing the exact changes made.
   */
  private Collection saveOnlyOld( Collection oldkeys, Map oldmap )
    throws Exception {
    
    Collection changes = new HashSet();

    for( Iterator it = oldkeys.iterator(); it.hasNext(); ) {
      ScriptSlotRef key = (ScriptSlotRef)it.next();
      SavedSlot savedslot = (SavedSlot)oldmap.get(key);
      

      ResRef restoredval = savedslot.delete(nwnmodule_.getModfile(), sezfile_);

      /* This slot has been restored to its previous value.
       */
      changes.add( new SlotValue( key, restoredval ) );
    }
    return changes;
  }



  /**
   * A new script slot in a blueprint has come under our purview. Install
   * the scripts and add a record to the ScriptEase .itp.
   *
   * @param newkeys ScriptSlotRef keys that index into newmap.
   * @param newmap ScriptSlot instances for our newly controlled slots.
   * @return Collection of SlotValue describing the exact changes made.
   */
  private Collection saveOnlyNew( Collection newkeys, Map newmap )
    throws Exception {
    Collection changes = new HashSet();

    for( Iterator it = newkeys.iterator(); it.hasNext(); ) {
      ScriptSlotRef key = (ScriptSlotRef)it.next();
      ScriptSlot scriptslot = (ScriptSlot)newmap.get(key);

      //System.out.println("about to save " + key);
      changes.add( saveOneOnlyNew( scriptslot ) );
    }
    return changes;
  }

  /**
   * @return A SlotValue summarizing the change.
   */
  private SlotValue saveOneOnlyNew( ScriptSlot scriptslot ) throws Exception {
    /* Need to add the script resource to the module file,
     * hook it into the script-attachable entity, and keep a record in
     * the scriptease ".sez" file.
     */
    
    /* So first assign a new name. 
     */
    String scriptresname = generateName();

    /* Put the scripts in the module.
     */
    ModFile modfile = nwnmodule_.getModfile();
    modfile.addupdateresource( new ResRef(scriptresname, ResRef.tiNSS),
			       scriptslot.getSourcecode().getBytes() );
    modfile.addupdateresource( new ResRef(scriptresname, ResRef.tiNCS),
			       scriptslot.getBytecode() );

    /* And save the info about the slot.
     */
    scriptslot.getSlotref().save( modfile, sezfile_,
				  new ResRef(scriptresname,-1) );

    return new SlotValue( scriptslot.getSlotref(),
			  new ResRef(scriptresname, -1) );
  }

  /**
   * A script slot in a script-attachable entity continues to be occupied,
   * but there is new content for the script.
   * @param bothkeys ScriptSlotRefs that key into both oldmap and newmap
   * @param oldmap Values are SavedSlot.
   * @param newmap Values are ScriptSlot
   * @return Collection of SlotValue describing the exact changes made.
   */
  private Collection saveBoth( Collection bothkeys, Map oldmap, Map newmap ) 
    throws Exception {

    for( Iterator it = bothkeys.iterator(); it.hasNext(); ) {
      ScriptSlotRef key = (ScriptSlotRef)it.next();
      SavedSlot savedslot = (SavedSlot)oldmap.get(key);
      ScriptSlot newslot= (ScriptSlot)newmap.get(key);

      
      savedslot.update( nwnmodule_.getModfile(), sezfile_,
			newslot.getCode() );


    }

    /* The slots point to the same script names, so essentially no changes
     * are made.
     */
    return new HashSet();
  }


  private String generateName() throws Exception {
    String base = "scriptease";
    ModFile mod = nwnmodule_.getModfile();
    for( int i = 0; i < 1000000000; ++i ) {
      String name = base + Integer.toString(i);
      if( !mod.contains( new ResRef(name, ResRef.tiNSS) ) &&
	  !mod.contains( new ResRef(name, ResRef.tiNCS) ) ) {
	// Good.
	return name;
      }
    }
    /* Um, that would have taken a long time.
     */
    throw new Exception("This module has a lot of scripts.");
  }


}
