/* -*- c-basic-offset: 2 -*- */

package nwnintf.format.itp;

import nwnintf.format.Prim;

import java.util.Vector;
import java.util.Enumeration;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * Represent an "itp" file. Also called "GFF".
 * The current implementation is according to Torlack's document:
 * http://www.torlack.com/index.html?topics=nwndata_itp
 *  which was superceded by:
 * http://nwn.bioware.com/developers/erf.html
 * So there may be errors and omissions in this code.
 */
public class ItpFile {

  // Byte offsets of elements.
  public static int SIZE           = 56;
  public static int TYPE           = 0;
  public static int VERSION        = 4;
  public static int ENTRYOFFSET    = 8;
  public static int NENTRIES       = 12;
  public static int ELEMENTOFFSET  = 16;
  public static int NELEMENTS      = 20;
  public static int VARNAMEOFFSET  = 24;
  public static int NVARNAMES      = 28;
  public static int VARDATAOFFSET  = 32;
  public static int NVARDATABYTES  = 36;
  public static int MULTIMAPOFFSET = 40;
  public static int NMULTIMAPBYTES = 44;
  public static int LISTOFFSET     = 48;
  public static int LISTBYTES      = 52;


  String type_, version_;
  Vector entries_, elements_;

  /**
   * Create a new ItpFile from scratch.
   */
  public ItpFile( String type, String version ) {
    type_ = type;
    version_ = version;

    entries_ = new Vector();
    elements_ = new Vector();
    
    // Initalize with one root entry.
    entries_.add(new ItpEntry(this));
  }

  /**
   * Interpret a sequence of bytes representing a GFF file.
   */ 
  public ItpFile( byte[] bytes ) throws Exception {
    this( bytes, 0 );
  }

  /**
   * Interpret a sequence of bytes starting at a particular offset
   * as a GFF File.
   */
  public ItpFile( byte[] bytes, int offset ) throws Exception {

    /* Separate allocation step before read because the ItpEntry and
     * ItpElements want handles to other entries and elements before
     * they're fully constructed.
     */

    type_ = Prim.string(bytes, offset + TYPE, 4);
    version_ = Prim.string( bytes, offset + VERSION, 4);

    allocateEntries( bytes, offset );
    allocateElements( bytes, offset );
    readEntries( bytes, offset );
    readElements( bytes, offset );
  }


  private void allocateEntries( byte[] bytes, int offset ) {
    int nentries = Prim.int32lsb( bytes, offset + NENTRIES );
    entries_ = new Vector();

    for( int i = 0; i < nentries; ++i ) {
      entries_.add( new ItpEntry( this ) );
    }
  }

  private void allocateElements( byte[] bytes, int offset ) throws
  Exception {
    int nelements = Prim.int32lsb( bytes, offset + NELEMENTS );
    int elementoffset = Prim.int32lsb( bytes, offset + ELEMENTOFFSET );
    elements_ = new Vector();

    for( int i =0; i < nelements; ++i ) {
      ItpElement elt = 
	ItpElement.newElement( this, bytes,
			       offset + elementoffset + i * ItpElement.SIZE );
      elements_.add( elt );
    }
  }



  private void readEntries( byte[] bytes, int fileoffset ) {
    int nentries = Prim.int32lsb( bytes, fileoffset + NENTRIES );
    int entryoffset = Prim.int32lsb( bytes, fileoffset + ENTRYOFFSET );
    int multimapoffset = Prim.int32lsb( bytes, fileoffset + MULTIMAPOFFSET );
    int elementoffset = Prim.int32lsb( bytes, fileoffset + ELEMENTOFFSET );

    for( int i = 0; i < nentries; ++i ) {
      ItpEntry entry = (ItpEntry)entries_.elementAt(i);
      entry.read( bytes, fileoffset + entryoffset + 
		  ItpEntry.SIZE * i, multimapoffset );
    }
  }

  private void readElements( byte[] bytes, int fileoffset ) {
    int nelements = Prim.int32lsb( bytes, fileoffset + NELEMENTS );
    int elementoffset = Prim.int32lsb( bytes, fileoffset + ELEMENTOFFSET );

    int varnameoffset = Prim.int32lsb( bytes, fileoffset + VARNAMEOFFSET );
    int vardataoffset = Prim.int32lsb( bytes, fileoffset + VARDATAOFFSET );
    int listoffset = Prim.int32lsb( bytes, fileoffset + LISTOFFSET );
    for( int i = 0; i < nelements; ++i ) {
      ((ItpElement)elements_.elementAt(i)).read( bytes,
						 fileoffset + elementoffset + 
						 ItpElement.SIZE * i,
						 varnameoffset, 
						 vardataoffset, listoffset
						 );
    }
    
  }


  /** Return the 
   */
  public ItpEntry entry(int idx) {
    return (ItpEntry)entries_.elementAt(idx);
  }
  public ItpElement element(int idx){
    return (ItpElement)elements_.elementAt(idx);
  }


  private static void dump(byte[] bytes, int fileoffset) {
    System.out.print("type/version ");
    System.out.println(Prim.string0( bytes, fileoffset + TYPE, 8 ));
    System.out.print("entryoffset ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + ENTRYOFFSET));
    System.out.print("nentries ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + NENTRIES));
    System.out.print("elementoffset ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + ELEMENTOFFSET));
    System.out.print("nelements ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + NELEMENTS));
    System.out.print("varnameoffset ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + VARNAMEOFFSET));
    System.out.print("nvarnames ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + NVARNAMES));
    System.out.print("vardataoffset ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + VARDATAOFFSET));
    System.out.print("nvardatabytes ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + NVARDATABYTES));
    System.out.print("multimapoffset ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + MULTIMAPOFFSET));
    System.out.print("nmultimapbytes ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + NMULTIMAPBYTES));
    System.out.print("listoffset ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + LISTOFFSET));
    System.out.print("listbytes ");
    System.out.println(Prim.int32lsb(bytes, fileoffset + LISTBYTES));
  }

  public ItpEntry root() {
    return entry(0);
  }

  public int indexOfEntry(ItpEntry ent) {
    for(int i = 0; i < entries_.size(); ++i ) {
      if( entries_.elementAt(i) == ent ) { return i; }
    }
    throw new IndexOutOfBoundsException("ItpEntry not found");
  }
  public int indexOfElement(ItpElement elt) {
    for(int i = 0; i < elements_.size(); ++i ) {
      if( elements_.elementAt(i) == elt ) { return i; }
    }
    throw new IndexOutOfBoundsException("ItpElement not found");
  }
  public int nentries() {
    return entries_.size();
  }
  public int nelements() {
    return elements_.size();
  }

  public void addEntry(ItpEntry entry) {
    entries_.add(entry);
  }
  public void addElement(ItpElement element) {
    elements_.add(element);
  }
  /**
   * Recursively deletes contained elements and entries.
   */
  public void removeEntry(ItpEntry entry) {
    entry.removeChildren();
    entries_.remove(entry);
  }

  /**
   * Recursively deletes contained elements and entries.
   */
  public void removeElement(ItpElement element) {
    element.removeChildren();
    elements_.remove(element);
  }

  public String type() {
    return type_;
  }
  public String version() {
    return version_;
  }

  public void setType(String type) {
    type_ = type;
  }
  public void setVersion(String version) {
    version_ = version;
  }

  /**
   * This test function may be out-of-date w.r.t. the API.
   */
  public static void main(String[] args) throws Exception {
    try {
      File f = new File(args[0]);

      int len = (int)f.length();

      byte b[] = new byte[len];

      FileInputStream fis = new FileInputStream(f);
      fis.read(b,0,(int)len);

      ItpFile.dump( b, 0 );
      ItpFile itpfile = new ItpFile( b, 0 );


      File g = new File(args[1]);
      FileOutputStream fos = new FileOutputStream(g);
      fos.write( new ItpWriter(itpfile).bytes() );
      fos.close();

    } catch( Error e ) {
      e.printStackTrace();
      throw e;
    }
  }

}
