/* -*- c-basic-offset: 2 -*- */

package nwnintf.format.mod;

import nwnintf.format.*;

import java.util.*;
import java.io.*;

/**
 * Represent a "mod"/"sav"/"erf"/"hak"/"nwm" etc. file. Cannot be used
 * to create a new module file entirely from scratch, but it could be
 * modified to do so.
 * See Torlack's file format description:
 * http://www.torlack.com/index.html?topics=nwndata_mod
 */
public class ModFile {

  // Byte offsets of elements in the header.
  public static int TYPE           = 0;
  public static int VERSION        = 4;
  public static int NSTRINGS       = 8;
  public static int STRINGSLEN     = 12;
  public static int NRESOURCES     = 16;
  public static int STRINGOFFSET   = 20;
  public static int RESOFFSET      = 24;
  public static int POSOFFSET      = 28;
  public static int YEAR           = 32;
  public static int DAY            = 36;
  public static int TYPECODE       = 40;
  public static int PADDING        = 44;
  public static int PADDINGSIZE    = 116;
  public static int SIZE           = 160;

  // Values for typecode
  public static int TYPE_MOD = -1;
  public static int TYPE_SAV = 0;
  public static int TYPE_NWM = -1;


  String type_, version_;

  int nstrings_;
  int stringslen_;
  /* Add strings_ when necessary.
   */
  //Langstring[] strings_;

  int nresources_;
  int stroffset_;
  int resoffset_;
  int posoffset_;

  int year_;
  int day_;
  int typecode_;
  
  Vector resdescs_;
  Vector resposns_;

  RandomAccessFile raf_;

  public ModFile( File f ) throws Exception {

    // Synchronous read-write mode because it has no explicit flush.
    //
    raf_ = new RandomAccessFile(f, "rws");

    type_ = Prim.string(raf_,4);
    version_ = Prim.string(raf_,4);
    nstrings_ = Prim.int32lsb(raf_);
    stringslen_ = Prim.int32lsb(raf_);

    nresources_ = Prim.int32lsb(raf_);
    stroffset_ = Prim.int32lsb(raf_);
    resoffset_ = Prim.int32lsb(raf_);
    posoffset_ = Prim.int32lsb(raf_);
    year_ = Prim.int32lsb(raf_);
    day_ = Prim.int32lsb(raf_);
    typecode_ = Prim.int32lsb(raf_);
    // Discard padding.
    Prim.string(raf_, 29 * 4);

    /* Finished reading header. Cache resource descriptions and positions
     * in memory.
     */
    raf_.seek( resoffset_ );
    resdescs_ = new Vector();
    for( int i = 0; i < nresources_; ++i ) {
      resdescs_.add( new ModResDesc( raf_ ) );
    }

    raf_.seek( posoffset_ );
    resposns_ = new Vector();
    for( int i = 0; i < nresources_; ++i ) {
      resposns_.add( new ModResPosn( raf_ ) );
    }

  }


  public String type() {
    return type_;
  }
  public String version() {
    return version_;
  }

  /**
   * Return the idx'th ModResDesc.
   */
  public ModResDesc descAt( int idx ) {
    return (ModResDesc)resdescs_.elementAt(idx);
  }
  public int nresources() {
    return nresources_;
  }

  /**
   * Produce an iterator that iterates over all ResRef objects in the
   * file that have the given type.
   */
  public Iterator iterator( int type ) {
    return new ResourceIterator(this,type);
  }


  /**
   * Retrieve index of resource, -1 if none.
   */
  private int resdescindex( ResRef resref ) {
    for( int i = 0; i < nresources_; ++i ) {

      ModResDesc desc = (ModResDesc) resdescs_.elementAt(i);

      if( resref.name().equals( desc.name() ) && 
	  resref.typeInt() == desc.type() ) {
	return i;
      }
    }
    return -1;
  }

  /**
   * Test whether the module contains a resource by this name and type.
   */
  public boolean contains( ResRef resref ) {
    return resdescindex( resref ) != -1;
  }

  /**
   * Get a copy of a resource file in the module.
   * Notes: this does a linear search through the list of resources every
   * time. Should cache a map of ResRef -> ModResDesc.
   * @return bytes of the resource, null if not found.
   */
  public byte[] resource( ResRef resref ) throws java.io.IOException {
    int descidx = resdescindex(resref);
    if( descidx == -1 ) return null;

    ModResDesc desc = (ModResDesc) resdescs_.elementAt(descidx);
    return resource(desc);
  }
  public byte[] resource( ModResDesc desc ) throws java.io.IOException {
    ModResPosn posn = (ModResPosn) resposns_.elementAt( desc.idx() );
    return resource(posn);
  }
  public byte[] resource( ModResPosn posn ) throws java.io.IOException {
    raf_.seek( posn.offset() );
    byte[] result = new byte[ posn.size() ];
    raf_.readFully(result);
	return result;
  }


  private void addnewresource( ResRef resref, byte[] resdata )
    throws java.io.IOException {

    /* A new resource. We move the resource description and position
       * arrays to make room, even though we may not really need to.
       * A future enhancement.
       */
    int newresoffset = (int)raf_.length();

    /* Copy the resource description array into memory.
       */
    byte[] olddescs = new byte[ ModResDesc.SIZE * nresources_ ];
    raf_.seek( resoffset_ );
    raf_.readFully( olddescs );

    /* Write it into its place at the end of the file.
       */
    raf_.seek( newresoffset );
    raf_.write( olddescs );

      /* Create the new resource description and write it.
       */
    ModResDesc newdesc = new ModResDesc( resref.name(), nresources_, resref.typeInt() );
    resdescs_.add( newdesc );

    Prim.writeBytes0( raf_, newdesc.name(), 16 );
    Prim.writeIntLSB( raf_, newdesc.idx() );
    Prim.writeIntLSB( raf_, newdesc.type() );

    /* Now move the position description array to the end of the file.
       */
    int newposoffset = (int)raf_.length();

    byte[] oldposns = new byte[ ModResPosn.SIZE * nresources_ ];
    raf_.seek( posoffset_ );
    raf_.readFully( oldposns );

    raf_.seek( newposoffset );
    raf_.write(oldposns);

    /* The resource data is going to go right after the new position
     * entry.
     */
    int resdataposn = (int)raf_.length() + ModResPosn.SIZE;

    /* Append the new position entry.
     */
    ModResPosn newposn = new ModResPosn( resdataposn, resdata.length );
    Prim.writeIntLSB( raf_, newposn.offset() );
    Prim.writeIntLSB( raf_, newposn.size() );
    resposns_.add( newposn );


      // Finally write the resource data.
    raf_.write( resdata );

    // Now update the file header.

      // Write the new offset of the resource array.
    raf_.seek( RESOFFSET );
    resoffset_ = newresoffset;
    Prim.writeIntLSB( raf_, resoffset_ );

    raf_.seek( POSOFFSET );
    posoffset_ = newposoffset;
    Prim.writeIntLSB( raf_, posoffset_ );

    raf_.seek( NRESOURCES );
    Prim.writeIntLSB( raf_, ++nresources_ );
  }


  /**
   * Add a new resource to the file, or update an existing one.
   * This function makes a mess of the file, wasting space and moving
   * things all over the place.
   * @param resref Name of resource file.
   * @param resdata Bytes making up the resource.
   *
   */
  public void addupdateresource( ResRef resref, byte[] resdata ) 
    throws java.io.IOException {
    int idx = resdescindex( resref );
    if( idx == -1 ) {

      // Totally new  resource.
      addnewresource( resref, resdata );

    } else {

      // We have a new version of an existing resource.
      
      ModResPosn posn = (ModResPosn)resposns_.elementAt(idx);

      if( posn.size() >= resdata.length ) {
	/* The new data is no longer than the current data, so it can
	 * be overwritten without moving anything.
	 */
	
	// Write the new resource data.
	raf_.seek( posn.offset() );
	raf_.write( resdata );

      } else {
	int offset = (int)raf_.length();
	raf_.seek(offset);
	raf_.write( resdata );
	posn.setOffset( offset );
      }

      /* In either case, update the resource length and offset,
       * (even if they didn't change).
       */
      posn.setSize( resdata.length );
      raf_.seek( posoffset_ + ModResPosn.SIZE * idx );
      Prim.writeIntLSB( raf_, posn.offset() ); // Even if it hadn't changed.
      Prim.writeIntLSB( raf_, posn.size() );
    }
    // There is no flush method. How lame.
    //raf_.flush();
  }


  /**
   * Delete a resource from the module file.
   */
  public void deleteresource( ResRef resref )
    throws java.io.IOException {
    int idx = resdescindex( resref );
    if( idx == -1 ) {
      throw new IndexOutOfBoundsException("Resource \"" + resref.filename() + "\" not found");
    }

    ModResDesc resdesc = (ModResDesc)resdescs_.elementAt(idx);
    int posidx = resdesc.idx();

    nresources_ --;
    raf_.seek(NRESOURCES);
    Prim.writeIntLSB(raf_,nresources_);
    resdescs_.remove(idx);
    resposns_.remove(posidx);

    // Adjust position indices in the resource descriptions downwards.
    for( int i = 0; i < nresources_; ++i ) {
      ModResDesc resi = (ModResDesc)resdescs_.elementAt(i);
      if( resi.idx() > posidx ) {
	resi.setIdx( resi.idx() - 1);
      }
    }

    // Rewrite the position and description arrays where they stand.
    raf_.seek( resoffset_ );
    for( int i = 0; i < nresources_; ++i ) {
      ModResDesc desc = (ModResDesc)resdescs_.elementAt(i);
      Prim.writeBytes0( raf_, desc.name(), 16 );
      Prim.writeIntLSB( raf_, desc.idx() );
      Prim.writeIntLSB( raf_, desc.type() );
    }
    raf_.seek( posoffset_ );
    for( int i = 0; i < nresources_; ++i ) {
      ModResPosn posn = (ModResPosn)resposns_.elementAt(i);
      Prim.writeIntLSB( raf_, posn.offset() );
      Prim.writeIntLSB( raf_, posn.size() );
    }

  }


  public static void main(String[] args) throws Exception {
    try {
      File f = new File(args[0]);

      ModFile mf = new ModFile(f);

      RandomAccessFile resf = new RandomAccessFile(args[1], "r");
      byte[] resdata = new byte[(int)resf.length()];
      resf.readFully(resdata);
      mf.addupdateresource( ResRef.parseFilename(args[1]), resdata );


      /*
      byte[] b = mf.resource( args[1] );
      File fout = new File(args[1]);

      FileOutputStream fos = new FileOutputStream(fout);
      fos.write( b );
      fos.close();
      */
      
      /*
      int len = (int)f.length();

      byte b[] = new byte[len];

      FileInputStream fis = new FileInputStream(f);
      fis.read(b,0,(int)len);

      ItpFile.dump( b, 0 );
      ItpFile itpfile = new ItpFile( b, 0 );


      File g = new File(args[1]);
      FileOutputStream fos = new FileOutputStream(g);
      fos.write( new ItpWriter(itpfile).bytes() );
      fos.close();
      */

    } catch( Error e ) {
      e.printStackTrace();
      throw e;
    }
  }
}

class ResourceIterator implements Iterator {
  private ModFile mod_;
  private int type_;
  private int idx_; // next index to examine.
  private int nresources_;

  private ModResDesc next_;
  private boolean nextavail_; // does next_ actually point to the next elt.

  ResourceIterator( ModFile mod, int type ) {
    mod_ = mod;
    type_ = type;
    idx_ = 0;
    nresources_ = mod_.nresources();
    next_ = null;
    nextavail_ = false;
    produceNext();
  }

  private boolean produceNext() {
    if( nextavail_ ) return true;

    while( idx_ < nresources_ ) {
      ModResDesc desc = mod_.descAt(idx_++);
      if( desc.type() == type_ ) {
	next_ = desc;
	nextavail_ = true;
	break;
      }
    }
    return nextavail_;
  }

  public boolean hasNext() {
    return produceNext();
  }

  /**
   * Returns ResRef of resource.
   */
  public Object next() throws NoSuchElementException {
    if( ! produceNext() ) throw new NoSuchElementException();
    nextavail_ = false;
    
    return next_.resRef();
  }

  public void remove() {
    throw new UnsupportedOperationException();
  }
    
}
